"""
Unit tests for SQLAlchemy utilities to ensure that we convert a str representation of a data type
to the correct sqlalchemy type
"""
import geoalchemy2
import pytest
import sqlalchemy as sa
import sqlalchemy.dialects.postgresql as sa_pg

from yalchemy import sqla_utils


@pytest.mark.parametrize('expected,col_type,col_args', [
    (sa.String(255), 'varchar', [255]),
    (sa.String(10), 'varchar', [10]),
    (sa.Text(), 'text', []),
    (sa.CHAR(), 'char', []),
    (sa_pg.BYTEA(), 'bytea', []),
    (sa_pg.BIT(varying=True), 'bit varying', []),
    (sa.Date(), 'date', []),
    (sa.TIMESTAMP(), 'datetime', []),
    (sa.TIMESTAMP(), 'datetime_without_time_zone', []),
    (sa.TIMESTAMP(timezone=True), 'datetime_with_time_zone', []),
    (sa.TIMESTAMP(), 'timestamp', []),
    (sa.TIMESTAMP(), 'timestamp_without_time_zone', []),
    (sa.TIMESTAMP(timezone=True), 'timestamp_with_time_zone', []),
    (sa.Time(), 'time', []),
    (sa.Time(), 'time', []),
    (sa.Time(timezone=True), 'time_with_time_zone', []),
    (sa.Boolean(), 'boolean', []),
    (sa.Integer(), 'integer', []),
    (sa.BigInteger(), 'bigint', []),
    (sa.SmallInteger(), 'smallint', []),
    (sa_pg.UUID(), 'uuid', []),
    (sa_pg.UUID(as_uuid=True), 'uuid', []),
    (sa_pg.JSON(), 'json', []),
    (sa_pg.JSONB(), 'jsonb', []),
    (sa.Numeric(10, 2), 'numeric', [10, 2]),
    (sa.Numeric(), 'numeric', []),
    (sa.REAL(), 'real', []),
    (sa_pg.DOUBLE_PRECISION(), 'double_precision', []),
    (sa_pg.INTERVAL(), 'interval', []),
    (sa_pg.TSTZRANGE(), 'tstzrange', []),
    (sa_pg.INT4RANGE(), 'int4range', []),
    (sa_pg.DATERANGE(), 'daterange', []),
    (geoalchemy2.Geography('point', 4326), 'geography', ['Point', 4326]),
    # arrays
    (sa_pg.ARRAY(sa.Integer), 'integer[]', []),
    # errors
    pytest.mark.xfail((sa.types.NullType(), 'N/A', []),
                      raises=sqla_utils.UnknownColumnType, strict=True),
    ], ids=str)
def test_sa_type_from_str(expected, col_type, col_args):
    sa_type = sqla_utils.sa_type_from_str(col_type, col_args)
    assert sa_type.compile(sa_pg.dialect()) == expected.compile(sa_pg.dialect())


def test_sa_type_arrays():
    """ Test that we handle arrays properly with sa types """

    sa_type = sqla_utils.sa_type_from_str('text[]')
    assert isinstance(sa_type, sa_pg.ARRAY)
    assert isinstance(sa_type.item_type, sa.Text)


def test_invalid_type():
    """ Test that we raise an UnknownType error when we
    put in some weird postgres type """

    with pytest.raises(sqla_utils.UnknownColumnType):
        sqla_utils.sa_type_from_str('unknown_type')



""" Module to make the yaml dumps
of yalchemy dicts prettier than if we
do just a normal dump with yaml.safe_dump """

import collections

import yaml


class PrettyYalchemyDumper(yaml.SafeDumper):  # pylint: disable=too-many-ancestors
    """ Our custom yaml dumper to increase indents
    and to also handle OrderedDict and print it's keys
    out in order """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_representer(collections.OrderedDict, self.dict_representer)

    def increase_indent(self, flow=False, indentless=False):
        """ Increase the indent by forcing indentless to be False """
        return super().increase_indent(flow, False)

    @staticmethod
    def dict_representer(dumper, data):
        return dumper.represent_dict(data.items())



#!/usr/bin/env python3

"""
Deploys the package to PyPI and uploads docs. This script is called in circle.yml.
"""
import os
import shlex
import subprocess
import sys

CIRCLECI_ENV_VAR = 'CIRCLECI'


def _shell(cmd, check=True, stdin=None, stdout=None, stderr=None):  # pragma: no cover
    """Runs a subprocess shell with check=True by default"""
    return subprocess.run(cmd, shell=True, check=check, stdin=stdin, stdout=stdout, stderr=stderr)


def _pypi_push(dist):
    """Push created package to PyPI.

    Requires the following defined environment variables:
        - TWINE_USERNAME: The PyPI username to upload this package under
        - TWINE_PASSWORD: The password to the user's account

    Args:
        dist (str):
            The distribution to push. Must be a valid directory; shell globs are
            NOT allowed.
    """
    # Register all distributions and wheels with PyPI. We have to list the dist
    # directory and register each file individually because `twine` doesn't
    # handle globs.
    for filename in os.listdir(dist):
        full_path = os.path.join(dist, filename)
        if os.path.isfile(full_path):
            # This will fail if the project has never been uploaded, so use check=false
            _shell('twine register ' + shlex.quote(full_path), check=False)

    _shell('twine upload ' + shlex.quote(dist + '/*'))


def deploy(target):
    """Deploys the package and documentation.

    Proceeds in the following steps:

    1. Ensures proper environment variables are set and checks that we are on Circle CI
    2. Tags the repository with the new version
    3. Creates a standard distribution and a wheel
    4. Updates version.py to have the proper version
    5. Commits the ChangeLog, AUTHORS, and version.py file
    6. Pushes to PyPI
    7. Pushes the tags and newly committed files

    Raises:
        `EnvironmentError`:
            - Not running on CircleCI
            - `*_PYPI_USERNAME` and/or `*_PYPI_PASSWORD` environment variables
               are missing
            - Attempting to deploy to production from a branch that isn't master
    """
    # Ensure proper environment
    if not os.getenv(CIRCLECI_ENV_VAR):     # pragma: no cover
        raise EnvironmentError('Must be on CircleCI to run this script')

    current_branch = os.getenv('CIRCLE_BRANCH')
    if (target == 'PROD') and (current_branch != 'master'):
        raise EnvironmentError((
            'Refusing to deploy to production from branch {current_branch!r}. '
            'Production deploys can only be made from master.'
        ).format(current_branch=current_branch))

    if target in ('PROD', 'TEST'):
        pypi_username = os.getenv('{target}_PYPI_USERNAME'.format(target=target))
        pypi_password = os.getenv('{target}_PYPI_PASSWORD'.format(target=target))
    else:
        raise ValueError(
            "Deploy target must be 'PROD' or 'TEST', got {target!r}.".format(target=target))

    if not (pypi_username and pypi_password):  # pragma: no cover
        raise EnvironmentError((
            "Missing '{target}_PYPI_USERNAME' and/or '{target}_PYPI_PASSWORD' "
            "environment variables. These are required to push to PyPI."
        ).format(target=target))

    # Twine requires these environment variables to be set. Subprocesses will
    # inherit these when we invoke them, so no need to pass them on the command
    # line. We want to avoid that in case something's logging each command run.
    os.environ['TWINE_USERNAME'] = pypi_username
    os.environ['TWINE_PASSWORD'] = pypi_password

    # Set up git on circle to push to the current branch
    _shell('git config --global user.email "oss@cloverhealth.com"')
    _shell('git config --global user.name "Circle CI"')
    _shell('git config push.default current')

    # Obtain the version to deploy
    ret = _shell('make version', stdout=subprocess.PIPE)
    version = ret.stdout.decode('utf-8').strip()

    print('Deploying version {version!r}...'.format(version=version))

    # Tag the version
    _shell('git tag -f -a {version} -m "Version {version}"'.format(version=version))

    # Update the version
    _shell(
        'sed -i.bak "s/^__version__ = .*/__version__ = {version!r}/" */version.py'.format(
            version=version))

    # Create a standard distribution and a wheel
    _shell('python setup.py sdist bdist_wheel')

    # Add the updated ChangeLog and AUTHORS
    _shell('git add ChangeLog AUTHORS */version.py')

    # Start the commit message with "Merge" so that PBR will ignore it in the
    # ChangeLog. Use [skip ci] to ensure CircleCI doesn't recursively deploy.
    _shell('git commit --no-verify -m "Merge autogenerated files [skip ci]"')

    # Push the distributions to PyPI.
    _pypi_push('dist')

    # Push the tag and AUTHORS / ChangeLog after successful PyPI deploy
    _shell('git push --follow-tags')

    print('Deployment complete. Latest version is {version}.'.format(version=version))


if __name__ == '__main__':  # pragma: no cover
    if len(sys.argv) != 2:
        raise RuntimeError('Require one argument indicating deploy target: `prod` or `test`.')

    deploy(sys.argv[1].upper())



"""
Copyright (c) 2005-2018 the SQLAlchemy authors and contributors <see AUTHORS file>.
SQLAlchemy is a trademark of Michael Bayer.

SQLAlchemy utility which have been modified for use with yalchemy
"""
import re
import sqlalchemy.dialects.postgresql as sa_pg


class UnknownColumnType(Exception):
    """ Raised when the type given for a column is
    not known to Sqlalchemy """


def sa_type_from_str(col_type, col_args=()):
    """Convert a str datatype into a sqlalchemy datatype.
    This is mostly a copy and paste from the _get_column_info
    located at sa_pg.dialect()._get_column_info() """

    dialect = sa_pg.dialect()

    # replace _ with spaces and uppercase
    col_type = col_type.replace('_', ' ').lower()

    # strip (*) from character varying(5), timestamp(5)
    # with time zone, geometry(POLYGON), etc.
    attype = re.sub(r'\(.*\)', '', col_type)

    # strip '[]' from integer[], etc.
    attype = attype.replace('[]', '')

    is_array = col_type.endswith('[]')
    attype, args, kwargs = _special_att_handling(attype, col_args)

    coltype = dialect.ischema_names.get(attype)
    if coltype:
        coltype = coltype(*args, **kwargs)
        if is_array:
            coltype = dialect.ischema_names['_array'](coltype)
    else:
        raise UnknownColumnType(
            "Did not recognize column type '%s'" % attype)
    return coltype


# pylint: disable=too-many-statements,too-many-branches
def _special_att_handling(attype, col_args):  # noqa: C901
    """ laundry list of special handling that sqlalchemy
    does to convert between a Postgres type and a Sqlalchemy type """

    kwargs = {}

    if attype == 'uuid':
        kwargs = {'as_uuid': True}
        args = ()
    elif attype == 'numeric':
        if col_args:
            prec, scale = col_args
            args = (int(prec), int(scale))
        else:
            args = ()
    elif attype == 'double precision':
        args = (53, )
    elif attype == 'integer':
        args = ()
    elif attype in ('timestamp with time zone',
                    'time with time zone',
                    'datetime with time zone'):
        attype = attype.replace('datetime', 'timestamp')
        kwargs['timezone'] = True
        args = ()
    elif attype in ('timestamp without time zone',
                    'time without time zone',
                    'datetime without time zone',
                    'datetime',
                    'time'):
        attype = attype.replace('datetime', 'timestamp')
        kwargs['timezone'] = False
        args = ()
    elif attype == 'bit varying':
        kwargs['varying'] = True
        if col_args:  # pragma: no cover
            assert len(col_args) == 1
            args = (col_args[0],)
        else:
            args = ()
    elif attype == 'varchar':
        attype = 'character varying'
        if col_args:
            assert len(col_args) == 1
            args = (col_args[0],)
        else:  # pragma: no cover
            args = ()
    elif attype == 'char':
        attype = 'character'
        args = col_args
    elif attype.startswith('interval'):
        field_match = re.match(r'interval (.+)', attype, re.I)
        if col_args:  # pragma: no cover
            assert len(col_args) == 1
            kwargs['precision'] = col_args[0]
        if field_match:  # pragma: no cover
            kwargs['fields'] = field_match.group(1)
        attype = "interval"
        args = ()
    else:
        args = col_args or ()
    return attype, args, kwargs



""" Helper utilities for converting to a sqlalchemy type
from a string """
# pylint: disable=arguments-differ

import re

POSTGRES_ALIASES = {
    'int8': 'bigint',
    'bool': 'boolean',
    'decimal': 'numeric',
    'float4': 'real',
    'float8': 'double precision',
    'int2': 'smallint',
    'int4': 'integer',
    'int': 'integer',
    'serial2': 'smallserial',
    'serial4': 'serial',
    'timetz': 'time with time zone',
    'timestamptz': 'timestamp with time zone'
}


REGEX_SCHEMA_QUALIFIED = re.compile(r'^\w+\.\w+$')


def clean_datatype(datatype):
    """ Datatypes can have aliases (such as int for integer). This cleans
    the datatypes so there is always one consistent datatype used for the
    Column.datatype module """

    raw_datatype = str(datatype).strip()

    is_array = datatype.endswith('[]')
    if is_array:
        raw_datatype = raw_datatype[:-2].strip()  # remove array and whitespace before array

    raw_datatype = POSTGRES_ALIASES.get(raw_datatype, raw_datatype)
    clean_datatype = re.sub(r'\W+', '_', raw_datatype)
    if is_array:
        return clean_datatype + '[]'
    else:
        return clean_datatype


def is_schema_qualified(expression):
    return REGEX_SCHEMA_QUALIFIED.fullmatch(expression) is not None


def unquote_sql_string(expression):
    return expression.strip("'\"")



""" Test that we print yaml out in a pretty way
with the keys ordered by the __slots__ order """


from yalchemy import table_structures


TEST_YAML = """
tables:
  - name: test_table
    schema: test_schema
    columns:
      - {name: my_col, datatype: text, required: false}
"""

REVERSE_TEST_YAML = """
tables:
  - columns:
      - {required: false, datatype: text, name: my_col}
    schema: test_schema
    name: test_table
"""


def test_pretty_yaml_printing():
    """ Test a simple case that we correctly
    orders a table and column by it's slots  """

    col = table_structures.Column('my_col', 'text')
    table = table_structures.Table(name='test_table', schema='test_schema', columns=[col])
    table_set = table_structures.TableSet(tables=[table])
    assert table_set.to_yaml() == TEST_YAML.lstrip()

    # test with reversing the __slots__ order in the table and column
    table.__slots__ = table.__slots__[::-1]
    col.__slots__ = col.__slots__[::-1]
    assert table_set.to_yaml() == REVERSE_TEST_YAML.lstrip()
